//
//  DietData.h
//  MoveCoachMobile
//
//  Created by Alex Morgan on 7/18/13.
//

#import <Foundation/Foundation.h>

/**
 *  This Class is part of the Data Model and represents a single instance of a user Dietary entry.  The UserData Class maintains an array of DietData objects to provide a history of user dietary entries.
 */

@interface DietData : NSObject<NSCoding>

/**
 *  This property represents the name of the food entered into the system
 */
@property (nonatomic, strong)NSString *foodName;

/**
 *  This property represents the database code of the food entered into the system
 */
@property (nonatomic, strong)NSString *foodCode;

/**
 *  This property represents the description of the portion size for the food entered into the system
 */
@property (nonatomic, strong)NSString *portionDesc;

/**
 *  This property represents the meal that the food entered into the system was consumed at
 */
@property (nonatomic, strong)NSString *meal;

/**
 *  This property represents the date the food was entered into the system as a string.
 *
 *  NOTE: use referenceDate for all computational, db, or locale-sensitive functions.
 */
@property (nonatomic, strong)NSString *dateCreated;

/**
 *  This property represents the date and time the food was entered into the system.
 *
 *  NOTE: use this proprerty for all computational, db, or locale-sensitive functions.
 */
@property (nonatomic, strong)NSDate *referenceDate;

/**
 *  This property represents the number of calories eaten for the food entered into the system
 */
@property (nonatomic)float caloriesEaten;

/**
 *  This property represents the multiplication factor for all nutrients based on the portion size for the food entered into the system for one serving.
 */
@property (nonatomic)float portionFactor;

/**
 *  This property represents the number of servings for the food entered into the system
 */
@property (nonatomic)float servings;

/**
 *  This method allows a DietData object to be created with a specific name, date, caloric content, food code, portion factor, portion description, number of servings, and meal (e.g., breakfast).
 *
 *  @param foodname - name of the food entered
 *
 *  @date - a specific date to set dateCreated to.
 *
 *  @param cal - number of calories from the food entered
 *
 *  @param foodCode - database food code for the food entered
 *
 *  @param portionfactor - multiplication factor based on the selected portion of the food entered
 *
 *  @param portiondesc - description of the portion size selected for the food entered
 *
 *  @param serv - number of servings of the food entered
 *
 *  @param mymeal - meal the selected food is entered for
 */
- (id)initWithFoodName:(NSString *)foodname date:(NSString *)date calories:(float)cal foodCode:(NSString *)foodcode portionFactor:(float)portionfactor portionDesc:(NSString *)portiondesc servings:(float)serv meal:(NSString *)mymeal;

/**
 *  This method allows a DietData object to be created with a specific name, date, caloric content, food code, portion factor, portion description, number of servings, and meal (e.g., breakfast, lunch, dinner, or snack).
 *
 *  @param foodname - name of the food entered
 *
 *  @date - a specific date to set dateCreated to.
 *
 *  @referenceDate - [optional] a specific NSDate to set referenceDate to if not passed, the date/time of instantiation will be used.
 *
 *  @param cal - number of calories from the food entered
 *
 *  @param foodCode - database food code for the food entered
 *
 *  @param portionfactor - multiplication factor based on the selected portion of the food entered
 *
 *  @param portiondesc - description of the portion size selected for the food entered
 *
 *  @param serv - number of servings of the food entered
 *
 *  @param mymeal - meal the selected food is entered for
 */
- (id)initWithFoodName:(NSString *)foodname date:(NSString *)date referenceDate:(NSDate *)refDate calories:(float)cal foodCode:(NSString *)foodcode portionFactor:(float)portionfactor portionDesc:(NSString *)portiondesc servings:(float)serv meal:(NSString *)mymeal;

/**
 *  This method will update the [[UserData dataStore] caloriesByDateDictionary] hash to subtract the calories associated with this objects meal from the hash
 *
 */
- (void) removeFromCalorieDictionary;

@end
